import gevent.monkey

from .services.thrift.ttypes import SquareException

gevent.monkey.patch_all()

from .exceptions import LineServiceException
from .e2ee import E2EE
from .cube import LineCube
from .helpers import Helpers
from .timelineBiz import TimelineBiz
from .timeline import Timeline
from .object import Object
from .poll import Poll
from .thrift import Thrift
from .api import API
from .config import Config
from .models import Models
from os import system


class CHRLINE(
    Models,
    Config,
    API,
    Thrift,
    Poll,
    Object,
    Timeline,
    TimelineBiz,
    Helpers,
    LineCube,
    E2EE,
):
    def __init__(
        self,
        authTokenOrEmail: str = None,
        password: str = None,
        device: str = "CHROMEOS",
        version: str = None,
        os_name: str = None,
        os_version: str = None,
        noLogin: bool = False,
        encType: int = 1,
        debug: bool = False,
        customDataId: str = None,
        phone: str = None,
        region: str = None,
        forwardedIp: str = None,
        useThrift: bool = False,
        forceTMCP: bool = False,
        savePath: str = None,
        proxy_string: str = None,
    ):
        r"""
        Line client for CHRLINE.

        Use authToken or Email & Password to Login
        phone + region to Login secondary devices (and Android).

        ------------------------
        device: `str`
            Line special device name, you can view and add in config.py.
        version: `str`
            The device's version. it may affect some functions.
        os_name: `str`
            Customized system OS name.
        os_version: `str`
            Customized system OS version.
        noLogin: `bool`
            Set whether not to login
        encType: `int`
            Encryption for requests.

            - 0:
                no encryption.
            - 1:
                legy encryption.
        debug: `bool`
            * Developer options *
            For view some params and logs
        customDataId: `str`
            Special the customData id
        forwardedIp: `str`
            Fake ip used to spoof the server.
            ** not necessarily work **
        useThrift: `bool`
            Set whether to use line thrift.
            If true, you must place line thrifts in `services\thrift`.
        forceTMCP: `bool`
            It will force the use of TMoreCompact protocol on TalkService.
        savePath: `str`
            Set base-save dir path.
        """
        self.encType = encType
        self.isDebug = debug
        self.customDataId = customDataId
        Models.__init__(self, savePath)
        Config.__init__(self, device)
        self.initAppConfig(device, version, os_name, os_version)
        API.__init__(self, forwardedIp, proxy_string)
        Thrift.__init__(self)
        self.is_login = False
        self.use_thrift = useThrift
        self.force_tmore = forceTMCP
        if region is not None:
            self.LINE_SERVICE_REGION = region

        if authTokenOrEmail is not None and password is not None:
            email_func = self.requestEmailLogin
            if device in self.TOKEN_V3_SUPPORT:
                email_func = self.requestEmailLoginV2
            email_func(authTokenOrEmail, password)
        elif authTokenOrEmail:
            self.authToken = authTokenOrEmail
        elif phone:
            self.requestPwlessLogin(phone, self.LINE_SERVICE_REGION)
        else:
            if not noLogin:
                sqr_func = self.requestSQR
                if device in self.TOKEN_V3_SUPPORT:
                    sqr_func = self.requestSQR2
                for b in sqr_func():
                    print(b)
        if self.authToken:
            self.initAll()

    def initAll(self):
        self.checkNextToken(False)
        self.profile = self.getProfile()
        __profile_err = self.checkAndGetValue(self.profile, "error")
        if __profile_err is not None:
            self.log(f"登入失敗... {__profile_err}")
            try:
                for b in self.requestSQR(False):
                    print(b)
            except:
                raise Exception(f"登入失敗... {__profile_err}")
            self.handleNextToken(b)
            return self.initAll()
        self.mid = self.checkAndGetValue(self.profile, "mid", 1)
        __displayName = self.checkAndGetValue(self.profile, "displayName", 20)
        self.log(f"[{__displayName}] 登入成功 ({self.mid}) / {self.DEVICE_TYPE}")
        if self.customDataId is None:
            self.customDataId = self.mid
        try:
            system(f"title CHRLINE - {__displayName}")
        except:
            pass
        self.revision = -1
        try:
            self.groups = self.checkAndGetValue(
                self.getAllChatMids(), "memberChatMids", 1
            )
        except Exception as e:
            self.log(f"[getAllChatMids] {e}")
            self.groups = []

        E2EE.__init__(self)
        Timeline.__init__(self)
        TimelineBiz.__init__(self)
        Poll.__init__(self)
        Object.__init__(self)
        LineCube.__init__(self)
        Helpers.__init__(self)

        self.is_login = True

        self.can_use_square = False
        self.squares = None
        try:
            _squares = self.getJoinedSquares()
            self.can_use_square = True
            self.squares = _squares
        except SquareException as e:
            self.log(f"Not support Square: {e.reason}")
        except LineServiceException as e:
            self.log(f"Not support Square: {e.message}")

        self.custom_data = {}
        self.getCustomData()



# -*- coding: utf-8 -*-
from .hksc.types import HookTypes
from .hksc.events import HookEvents
from .hksc.utility import HookUtility
from .hksc.database import SqliteDatabase, JsonDatabase
import threading
import time


class HooksTracer(HookTypes, HookUtility):
    HooksType = {"Operation": 0, "Content": 1, "Command": 2, "SquareEvent": 7}

    def __init__(self, cl, db=None, prefixes=["!"], accounts=[], db_type=3):
        self.cl = cl
        self.db = None
        if db_type == 2:
            self.db = SqliteDatabase(cl, db)
        elif db_type == 3:
            self.db = JsonDatabase(cl, db)
        else:
            raise NotImplementedError()
        self.prefixes = prefixes
        self.opFuncs = []
        self.contFuncs = []
        self.cmdFuncs = []
        self.beforeFuncs = []
        self.afterFuncs = []
        self.eventFuncs = HookEvents()
        self.accounts = accounts

        # OpenChat
        self.seFuncs = []

    def run(self, fetchType: int = 0, **kwargs):
        self.eventFuncs.onReady()
        for _cl in self.accounts:
            _td = threading.Thread(
                target=self.runByClient, args=(_cl, fetchType), kwargs=kwargs
            )
            _td.daemon = True
            _td.start()
        self.runByClient(self.cl, fetchType, **kwargs)

    def runByClient(self, cl, fetchType: int = 0, **kwargs):
        IGNORE_ERRORS = ["ConnectionResetError", "gaierror", "ConnectionError"]
        while cl.is_login:
            if fetchType == 0:
                for op in cl._Poll__fetchOps():
                    _td = threading.Thread(
                        target=self.trace, args=(op, self.HooksType["Operation"], cl)
                    )
                    _td.daemon = True
                    _td.start()
            elif fetchType == 2:
                cl.legyPushers.hook_callback = self.PushCallback
                if cl.DEVICE_TYPE not in ["ANDROID", "IOS", "DESKTOPWIN", "DESKTOPMAC", "IOSIPAD"]:
                    raise ValueError(f"device not supported PUSH: {cl.DEVICE_TYPE}")
                try:
                    cl.legyPushers.conns = []
                    cl.legyPushers.initializeConn()
                    self.eventFuncs.onInitializePushConn()
                    cl.legyPushers.InitAndRead(**kwargs)
                except Exception as e:
                    cl.log(f"[POLLING] Push dead by `{e.__class__.__name__}`: {e}")
                    if e.__class__.__name__ not in IGNORE_ERRORS:
                        raise e
                cl.log("[POLLING] Push conn is dead! retry after 3.3 sec.")
                time.sleep(3.3)
            else:
                raise ValueError("Invalid fetchType: %d" % fetchType)

    def trace(self, data, type, cl, *attr):
        if type == self.HooksType["Operation"]:
            _a = self.opFuncs
        elif type == self.HooksType["Content"]:
            _a = self.contFuncs
        elif type == self.HooksType["Command"]:
            _a = self.cmdFuncs
        elif type == self.HooksType["SquareEvent"]:
            _a = self.seFuncs
        else:
            raise Exception(f"unknow type: {type}")
        _b = self.beforeFuncs
        _c = self.afterFuncs
        for _before in _b:
            if _before.type == type:
                if _before(self, data, cl, *attr):
                    break
        for _func in _a:
            if _func(self, data, cl, *attr):
                return True
        for _after in _c:
            if _after.type == type:
                if _after(self, data, cl, *attr):
                    break
        return False

    def PushCallback(self, cl, serviceType, event):
        ht = None
        if serviceType == 3:
            ht = "SquareEvent"
        elif serviceType == 5:
            ht = "Operation"
        else:
            raise ValueError(f"Invalid serviceType: {serviceType}")
        _td = threading.Thread(target=self.trace, args=(event, self.HooksType[ht], cl))
        _td.daemon = True
        _td.start()



# -*- coding: utf-8 -*-

class PwlessPrimaryRegistrationService(object):
    
    def __init__(self):
        pass

    def createSession(self):
        params = []
        sqrd = self.generateDummyProtocol('createSession', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_PWLESS_PRIMARY_REGISTRATION_ENDPOINT ,sqrd, 4)

    def getChallengeForPrimaryReg(self, session):
        params = [
            [12, 1, [
                [11, 1, session]
            ]]
        ]
        sqrd = self.generateDummyProtocol('getChallengeForPrimaryReg', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_PWLESS_PRIMARY_REGISTRATION_ENDPOINT ,sqrd, 4)

    def registerPrimaryCredential(self, session: str, cId: str, cType: str):
        raise NotImplementedError('RegisterPrimaryCredential id not implemented.')
        params = [
            [12, 1, [
                [11, 1, session],
                [12, 2, [
                    [11, 1, cId],
                    [11, 2, cType],
                    # [12, 3, response],
                    # [12, 4, extensionResults]
                ]]
            ]]
        ]
        sqrd = self.generateDummyProtocol('registerPrimaryCredential', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_PWLESS_PRIMARY_REGISTRATION_ENDPOINT ,sqrd, 4, access_token=session)


# -*- coding: utf-8 -*-
import time
import json

class LiffService(object):
    LIFF_REQ_TYPE = 4
    LIFF_RES_TYPE = 4

    def __init__(self):
        pass
        
    def issueLiffView(self, chatMid, liffId="1562242036-RW04okm", lang='th_TH', deviceSetting=None):
        context = [12, 1, []]
        if chatMid is not None:
            chat = [11, 1, chatMid]
            if chatMid[0] in ['u', 'c', 'r']:
                chatType = 2
            else:
                chatType = 3
            context = [12, chatType, [chat]]
        params = [
            [12, 1, [
                [11, 1, liffId],
                [12, 2, [
                    context
                ]],
                [11, 3, lang],
                # [12, 4, []], # deviceSetting
                # [11, 5, None] # msit
            ]]
        ]
        sqrd = self.generateDummyProtocol('issueLiffView', params, self.LIFF_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.LINE_LIFF_ENDPOINT, sqrd, self.LIFF_RES_TYPE)
        
    def getLiffViewWithoutUserContext(self, liffId="1562242036-RW04okm"):
        b = self.TCompactProtocol()
        sqrd = [130, 33, 00] + self.getStringBytes('getLiffViewWithoutUserContext', isCompact=True)
        sqrd += b.getFieldHeader(12, 1)
        c = self.TCompactProtocol()
        sqrd += c.getFieldHeader(8, 1)
        sqrd += self.getStringBytes(liffId, isCompact=True)
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_LIFF_ENDPOINT ,sqrd, ttype=4)
        
    def issueSubLiffView(self, chatMid, msit, liffId="1562242036-RW04okm", lang='th_TH', deviceSetting=None):
        b = self.TCompactProtocol()
        sqrd = [130, 33, 00] + self.getStringBytes('issueSubLiffView', isCompact=True)
        sqrd += b.getFieldHeader(12, 1)
        c = self.TCompactProtocol()
        sqrd += c.getFieldHeader(8, 1)
        sqrd += self.getStringBytes(liffId, isCompact=True)
        sqrd += c.getFieldHeader(8, 3) + self.getStringBytes(lang, isCompact=True)
        sqrd += c.getFieldHeader(8, 5) + self.getStringBytes(msit, isCompact=True)
        sqrd += c.getFieldHeader(12, 2)
        d = self.TCompactProtocol()
        if chatMid[0] not in ['u', 'c', 'r']:
            sqrd += d.getFieldHeader(12, 3)
        else:
            sqrd += d.getFieldHeader(12, 2)
        e = self.TCompactProtocol()
        sqrd += e.getFieldHeader(8, 1)
        sqrd += self.getStringBytes(chatMid, isCompact=True)
        sqrd += [0, 0, 0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_LIFF_ENDPOINT ,sqrd, ttype=4)


from functools import wraps
import traceback,json
import re

class HookTypes(object):

    def Event(self, func):
        setattr(self.eventFuncs, func.__name__, func)
        return func

    def Operation(self, type: int):
        def __wrapper(func):
            @wraps(func)
            def __check(self, *args):
                op = args[0]
                opType = self.cl.checkAndGetValue(op, 'type', 3)
                if opType in [25, 26]:
                    message = self.cl.checkAndGetValue(op, 'message', 20)
                    message = self.cl.checkAndSetValue(message, 'opType', opType)
                    isE2EE = bool(self.cl.checkAndGetValue(message, 'chunks', 20))
                    message = self.cl.checkAndSetValue(message, 'isE2EE', isE2EE)
                    op = self.cl.checkAndSetValue(op, 'message', 20, message)
                if opType == type:
                    func(self, *args)
                    return True
                return False
            self.opFuncs.append(__check)
        return __wrapper

    def Content(self, type: int):
        def __wrapper(func):
            @wraps(func)
            def __check(self, *args):
                op = args[0]
                contentType = self.cl.checkAndGetValue(op, 'contentType', 15)
                if contentType == type:
                    func(self, *args)
                    return True
                return False
            self.contFuncs.append(__check)
        return __wrapper

    def Command(self, permissions: list = [], alt: list = [], toType: list=[0, 1, 2], ignoreCase: bool = False, inpart: bool = False, prefixes: bool = True, splitchar: str=None):
        """
        - permissions:
            List of permission's name.
            eg. ['admin']
        - alt:
            Sub cmds.
            eg: Command name is "GetOps", and use alt=["ops"], so u can use "GetOps" and "ops" to run the command.
        - toType:
            Only run when receives the command with the specific toType 
        - ignoreCase:
            *As the name suggests*
        - inpart:
            If False, match the command that same style.
            If True, match any command that starts with command name.
        - prefixes:
            Match the prefixs.
        - splitchar:
            If not None, match the first value after the command is split by this value
        """
        def __wrapper(func):
            func.prefixes = prefixes
            func.permissions = permissions
            func.toType = toType
            @wraps(func)
            def __check(self, *args):
                _fname = lambda _name = None: func.__name__ if _name is None else _name
                msg = args[0]
                sender = self.cl.checkAndGetValue(msg, '_from', 1)
                receiver = self.cl.checkAndGetValue(msg, 'to', 2)
                cl = args[1]
                if sender != cl.mid and not self.checkPermissions(sender, permissions):
                    return False
                msgToType = self.cl.checkAndGetValue(msg, 'toType', 3)
                msgType = self.cl.checkAndGetValue(msg, 'contentType', 15)
                isInpart = inpart
                if msgToType in toType:
                    if msgType == 0:
                        text = self.cl.checkAndGetValue(msg, 'text', 10)
                        for __name in [None] + alt:
                            fname = _fname(__name)
                            if text is None:
                                # TODO: E2EE Message
                                msg = self.cl.checkAndSetValue(msg, 'isE2EE', True)
                                text = cl.decryptE2EETextMessage(msg, self.cl.checkAndGetValue(msg, 'opType') == 25)
                                msg = self.cl.checkAndSetValue(msg, 'text', 10, text)  # maybe fixed in Operation__check?
                            if ignoreCase:
                                text = text.lower()
                                fname = fname.lower()
                            if splitchar is not None:
                                text = text.split(splitchar)[0]
                                isInpart = False # force
                            isUsePrefixes = not prefixes
                            if prefixes and len(self.prefixes) > 0:
                                for _prefix in self.prefixes:
                                    if text.startswith(_prefix):
                                        fname = _prefix + fname
                                        isUsePrefixes = True
                                        break
                                if not isUsePrefixes:
                                    return False
                            if text == fname or isInpart and text.startswith(fname):
                                func(self, *args)
                                return True
                    else:
                        raise ValueError(f'wrong content type: {msgType}')
                return False
            self.cmdFuncs.append(__check)
        return __wrapper

    def SquareEvent(self, SquareEventType: int):
        def __wrapper(func):
            @wraps(func)
            def __check(self, *args):
                event = args[0]
                _type = self.cl.checkAndGetValue(event, 'type', 3)
                if _type == SquareEventType:
                    func(self, *args)
                    return True
                return False
            self.seFuncs.append(__check)
        return __wrapper

    def Before(self, type: int):
        def __wrapper(func):
            func.type = type
            @wraps(func)
            def __bf(self, *args):
                func(self, *args)
                return True
            self.beforeFuncs.append(__bf)
        return __wrapper

    def After(self,type):
        def __wrapper(func):
            func.type = type
            @wraps(func)
            def __af(self, *args):
                func(self, *args)
                return True
            self.afterFuncs.append(__af)
        return __wrapper


# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, List
from .BaseService import BaseService, BaseServiceStruct

if TYPE_CHECKING:
    from CHRLINE import CHRLINE


class SquareService(BaseService):
    SquareService_REQ_TYPE = 4
    SquareService_RES_TYPE = 4
    SquareService_API_PATH = "/SQS1"

    SQUARE_EXCEPTION = {"code": 1, "message": 3, "metadata": 2}

    def __init__(self):
        self.SquareService_API_PATH = self.LINE_SQUARE_ENDPOINT

    def inviteIntoSquareChat(self, inviteeMids: list, squareChatMid: str):
        """Invite into square chat."""
        METHOD_NAME = "inviteIntoSquareChat"
        params = [[15, 1, [11, inviteeMids]], [11, 2, squareChatMid]]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def inviteToSquare(self, squareMid: str, invitees: list, squareChatMid: str):
        """Invite to square."""
        METHOD_NAME = "inviteToSquare"
        params = [[11, 2, squareMid], [15, 3, [11, invitees]], [11, 4, squareChatMid]]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def getJoinedSquares(self, continuationToken: str = None, limit: int = 50):
        """Get joined squares."""
        METHOD_NAME = "getJoinedSquares"
        params = [[11, 2, continuationToken], [8, 3, limit]]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def markAsRead(self, squareChatMid: str, messageId: str):
        """Mark as read for square chat."""
        METHOD_NAME = "markAsRead"
        params = [[11, 2, squareChatMid], [11, 4, messageId]]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def reactToMessage(self, squareChatMid: str, messageId: str, reactionType: int = 2):
        """
        React to message for square chat message.

        - reactionType
            ALL     = 0,
            UNDO    = 1,
            NICE    = 2,
            LOVE    = 3,
            FUN     = 4,
            AMAZING = 5,
            SAD     = 6,
            OMG     = 7,
        """
        METHOD_NAME = "reactToMessage"
        params = [
            [8, 1, 0],  # reqSeq
            [11, 2, squareChatMid],
            [11, 3, messageId],
            [8, 4, reactionType],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def findSquareByInvitationTicket(self, invitationTicket: str):
        """Find square by invitation ticket."""
        METHOD_NAME = "findSquareByInvitationTicket"
        params = [[11, 2, invitationTicket]]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def fetchMyEvents(
        self,
        subscriptionId: int = 0,
        syncToken: str = None,
        continuationToken: str = None,
        limit: int = 100,
    ):
        """Fetch square events."""
        METHOD_NAME = "fetchMyEvents"
        params = [
            [10, 1, subscriptionId],
            [11, 2, syncToken],
            [8, 3, limit],
            [11, 4, continuationToken],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def fetchSquareChatEvents(
        self,
        squareChatMid: str,
        syncToken: str = None,
        continuationToken: str = None,
        subscriptionId: int = 0,
        limit: int = 100,
    ):
        """Fetch square chat events."""
        METHOD_NAME = "fetchSquareChatEvents"
        fetchType = 1  # DEFAULT(1),PREFETCH_BY_SERVER(2),
        # PREFETCH_BY_CLIENT(3);
        params = [
            [10, 1, subscriptionId],
            [11, 2, squareChatMid],
            [11, 3, syncToken],
            [8, 4, limit],
            [8, 5, 1],  # direction
            [8, 6, 1],  # inclusive
            [11, 7, continuationToken],
            [8, 8, fetchType],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def sendSquareMessage(
        self,
        squareChatMid: str,
        text: str,
        contentType: int = 0,
        contentMetadata: dict = {},
        relatedMessageId: str = None,
    ):
        """Send message for square chat (OLD)."""
        METHOD_NAME = "sendMessage"
        message = [
            # [11, 1, _from],
            [11, 2, squareChatMid],
            [11, 10, text],
            [8, 15, contentType],  # contentType
            [13, 18, [11, 11, contentMetadata]],
        ]
        if relatedMessageId is not None:
            message.append([11, 21, relatedMessageId])
            message.append(
                [
                    8,
                    22,
                    3,
                ]
            )
            message.append([8, 24, 2])
        params = [
            [8, 1, self.getCurrReqId()],
            [11, 2, squareChatMid],
            [
                12,
                3,
                [
                    [12, 1, message],
                    [8, 3, 4],
                ],
            ],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def sendSquareTextMessage(
        self,
        squareChatMid: str,
        text: str,
        contentMetadata: dict = {},
        relatedMessageId: str = None,
    ):
        return self.sendSquareMessage(
            squareChatMid, text, 0, contentMetadata, relatedMessageId
        )

    def getSquare(self, squareMid: str):
        """Get square."""
        METHOD_NAME = "getSquare"
        params = [[11, 2, squareMid]]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def getJoinableSquareChats(
        self, squareMid: str, continuationToken: str = None, limit: int = 100
    ):
        """Get joinable square chats."""
        METHOD_NAME = "getJoinableSquareChats"
        params = [[11, 1, squareMid], [11, 10, continuationToken], [8, 11, limit]]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def createSquare(
        self,
        name: str,
        displayName: str,
        profileImageObsHash: str = "0h6tJf0hQsaVt3H0eLAsAWDFheczgHd3wTCTx2eApNKSoefHNVGRdwfgxbdgUMLi8MSngnPFMeNmpbLi8MSngnPFMeNmpbLi8MSngnOA",
        desc: str = "test with CHRLINE API",
        searchable: bool = True,
        SquareJoinMethodType: int = 0,
    ):
        """
        Create square.

        - SquareJoinMethodType
            NONE(0),
            APPROVAL(1),
            CODE(2);
        """
        METHOD_NAME = "createSquare"
        params = [
            [8, 2, self.getCurrReqId()],
            [
                12,
                2,
                [
                    [11, 2, name],
                    [11, 4, profileImageObsHash],
                    [11, 5, desc],
                    [2, 6, searchable],
                    [8, 7, 1],  # type
                    [8, 8, 1],  # categoryId
                    [10, 10, 0],  # revision
                    [2, 11, True],  # ableToUseInvitationTicket
                    [12, 14, [[8, 1, SquareJoinMethodType]]],
                    [2, 15, False],  # adultOnly
                    [15, 16, [11, []]],  # svcTags
                ],
            ],
            [
                12,
                3,
                [
                    [11, 3, displayName],
                    # [11, 4, profileImageObsHash],
                    [2, 5, True],  # ableToReceiveMessage
                    [10, 9, 0],  # revision
                ],
            ],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def getSquareChatAnnouncements(self, squareMid: str):
        """Get square chat announcements."""
        METHOD_NAME = "getSquareChatAnnouncements"
        params = [
            [11, 2, squareMid],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def leaveSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("leaveSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "leaveSquareChat", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareChatMember(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareChatMember is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareChatMember", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def searchSquares(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("searchSquares is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "searchSquares", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def updateSquareFeatureSet(self, updateAttributes: List[int], squareMid: str, revision: int, creatingSecretSquareChat: int = 0):
        """
        Update square feature set.

        - updateAttributes:
            CREATING_SECRET_SQUARE_CHAT(1),
            INVITING_INTO_OPEN_SQUARE_CHAT(2),
            CREATING_SQUARE_CHAT(3),
            READONLY_DEFAULT_CHAT(4),
            SHOWING_ADVERTISEMENT(5),
            DELEGATE_JOIN_TO_PLUG(6),
            DELEGATE_KICK_OUT_TO_PLUG(7),
            DISABLE_UPDATE_JOIN_METHOD(8),
            DISABLE_TRANSFER_ADMIN(9),
            CREATING_LIVE_TALK(10);
        """
        METHOD_NAME = "updateSquareFeatureSet"
        SquareFeatureSet = [
            [11, 1, squareMid],
            [10, 2, revision],
        ]
        if creatingSecretSquareChat != 0:
            SquareFeatureSet.append([12, 11, [
                [8, 1, 1],
                [8, 2, creatingSecretSquareChat]
            ]])
        params = [
            [14, 2, [8, updateAttributes]],
            [12, 3, SquareFeatureSet]
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def joinSquare(
        self,
        squareMid,
        displayName,
        ableToReceiveMessage: bool = False,
        passCode: str = None,
    ):
        params = [
            [
                12,
                1,
                [
                    [11, 2, squareMid],
                    [
                        12,
                        3,
                        [
                            [11, 2, squareMid],
                            [11, 3, displayName],
                            [2, 5, ableToReceiveMessage],
                            [10, 9, 0],
                        ],
                    ],
                    [12, 5, [[12, 2, [[11, 1, passCode]]]]],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("joinSquare", params, 4)
        return self.postPackDataAndGetUnpackRespData(
            self.LINE_SQUARE_ENDPOINT,
            sqrd,
            4,
            encType=0,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquarePopularKeywords(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getPopularKeywords is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getPopularKeywords", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def reportSquareMessage(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("reportSquareMessage is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "reportSquareMessage", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def updateSquareMemberRelation(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareMemberRelation is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareMemberRelation", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def leaveSquare(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("leaveSquare is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "leaveSquare", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareMemberRelations(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareMemberRelations is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareMemberRelations", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def removeSquareSubscriptions(self, subscriptionIds: list = []):
        params = [
            [
                12,
                1,
                [
                    [15, 2, [10, subscriptionIds]],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("removeSubscriptions", params, 4)
        return self.postPackDataAndGetUnpackRespData(
            self.LINE_SQUARE_ENDPOINT,
            sqrd,
            4,
            encType=0,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareMembers", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def updateSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareChat", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareMessageReactions(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getMessageReactions is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getMessageReactions", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def destroySquareMessage(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("destroyMessage is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "destroyMessage", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def reportSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("reportSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "reportSquareChat", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def unsendSquareMessage(self, squareChatMid: str, messageId: str):
        """
        Unsend message for square.

        2022/09/19: Added.
        """
        METHOD_NAME = "unsendMessage"
        params = SquareServiceStruct.UnsendMessageRequest(squareChatMid, messageId)
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
            readWith=f"{__class__.__name__}.{METHOD_NAME}",
        )

    def deleteSquareChatAnnouncement(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("deleteSquareChatAnnouncement is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "deleteSquareChatAnnouncement", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def createSquareChat(
        self,
        squareChatMid: str,
        name: str,
        chatImageObsHash: str,
        squareChatType: int = 1,
        maxMemberCount: int = 5000,
        ableToSearchMessage: int = 1,
        squareMemberMids: list = [],
    ):
        """
        - SquareChatType:
            OPEN(1),
            SECRET(2),
            ONE_ON_ONE(3),
            SQUARE_DEFAULT(4);
        - ableToSearchMessage:
            NONE(0),
            OFF(1),
            ON(2);
        """
        params = [
            [
                12,
                1,
                [
                    [8, 1, self.getCurrReqId()],
                    [
                        12,
                        2,
                        [
                            [11, 1, squareChatMid],
                            [8, 3, squareChatType],
                            [11, 4, name],
                            [11, 5, chatImageObsHash],
                            [8, 7, maxMemberCount],
                            [8, 11, ableToSearchMessage],
                        ],
                    ],
                    [15, 3, [11, squareMemberMids]],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol(
            "createSquareChat", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def deleteSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("deleteSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "deleteSquareChat", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareChatMembers(
        self, squareChatMid: str, continuationToken: str = None, limit: int = 200
    ):
        GetSquareChatMembersRequest = [[11, 1, squareChatMid], [8, 3, limit]]
        if continuationToken is not None:
            GetSquareChatMembersRequest.append([11, 2, continuationToken])
        params = [[12, 1, GetSquareChatMembersRequest]]
        sqrd = self.generateDummyProtocol(
            "getSquareChatMembers", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareFeatureSet(self, squareMid: str):
        """Get square feature set."""
        METHOD_NAME = "getSquareFeatureSet"
        params = [
            [11, 2, squareMid],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def updateSquareAuthority(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareAuthority is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareAuthority", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def rejectSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("rejectSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "rejectSquareMembers", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def deleteSquare(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("deleteSquare is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "deleteSquare", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def reportSquare(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("reportSquare is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "reportSquare", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareInvitationTicketUrl(self, mid: str):
        params = [
            [
                12,
                1,
                [
                    [11, 2, mid],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("getInvitationTicketUrl", params, 4)
        return self.postPackDataAndGetUnpackRespData(
            self.LINE_SQUARE_ENDPOINT,
            sqrd,
            4,
            encType=0,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def updateSquareChatMember(
        self,
        squareMemberMid: str,
        squareChatMid: str,
        notificationForMessage: bool = True,
        notificationForNewMember: bool = True,
        updatedAttrs: list = [6],
    ):
        """
        - SquareChatMemberAttribute:
            MEMBERSHIP_STATE(4),
            NOTIFICATION_MESSAGE(6),
            NOTIFICATION_NEW_MEMBER(7);
        """
        params = [
            [
                12,
                1,
                [
                    [14, 2, [8, updatedAttrs]],
                    [
                        12,
                        3,
                        [
                            [11, 1, squareMemberMid],
                            [11, 2, squareChatMid],
                            [2, 5, notificationForMessage],
                            [2, 6, notificationForNewMember],
                        ],
                    ],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("updateSquareChatMember", params, 4)
        return self.postPackDataAndGetUnpackRespData(
            self.LINE_SQUARE_ENDPOINT,
            sqrd,
            4,
            encType=0,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def updateSquareMember(
        self,
        updatedAttrs: list,
        updatedPreferenceAttrs: list,
        squareMemberMid: str,
        squareMid: str,
        revision: int,
        displayName: str = None,
        membershipState: int = None,
        role: int = None,
    ):
        """
        Update square member.

        SquareMemberAttribute:
            DISPLAY_NAME(1),
            PROFILE_IMAGE(2),
            ABLE_TO_RECEIVE_MESSAGE(3),
            MEMBERSHIP_STATE(5),
            ROLE(6),
            PREFERENCE(7);
        SquareMembershipState:
            JOIN_REQUESTED(1),
            JOINED(2),
            REJECTED(3),
            LEFT(4),
            KICK_OUT(5),
            BANNED(6),
            DELETED(7);
        """
        METHOD_NAME = "updateSquareMember"
        squareMember = [[11, 1, squareMemberMid], [11, 2, squareMid]]
        if 1 in updatedAttrs:
            if displayName is None:
                raise ValueError("displayName is None")
            squareMember.append([11, 3, displayName])
        if 5 in updatedAttrs:
            if membershipState is None:
                raise ValueError("membershipState is None")
            squareMember.append([8, 7, membershipState])
        if 6 in updatedAttrs:
            if role is None:
                raise ValueError("role is None")
            squareMember.append([8, 8, role])
        squareMember.append([10, 9, revision])
        params = [
            [14, 2, [8, updatedAttrs]],
            [14, 3, [8, updatedPreferenceAttrs]],
            [12, 4, squareMember],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    # https://discord.com/channels/466066749440393216/1076212330620256296
    def deleteOtherFromSquare(self, sid: str, pid: str):
        """Kick out member for square."""
        UPDATE_PREF_ATTRS = []
        UPDATE_ATTRS = [5]
        MEMBERSHIP_STATE = 5
        getSquareMemberResp = self.getSquareMember(pid)
        squareMember = self.checkAndGetValue(getSquareMemberResp, "squareMember", 1)
        squareMemberRevision = self.checkAndGetValue(squareMember, "revision", 9)
        revision = squareMemberRevision
        self.updateSquareMember(
            UPDATE_ATTRS,
            UPDATE_PREF_ATTRS,
            pid,
            sid,
            revision,
            membershipState=MEMBERSHIP_STATE,
        )

    def updateSquare(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquare is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquare", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareAuthorities(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareAuthorities is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareAuthorities", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def updateSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareMembers", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareChatStatus(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareChatStatus is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareChatStatus", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def approveSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("approveSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "approveSquareMembers", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareStatus(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareStatus is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareStatus", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def searchSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("searchSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "searchSquareMembers", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def checkSquareJoinCode(self, squareMid: str, code: str):
        params = [
            [
                12,
                1,
                [
                    [11, 2, squareMid],
                    [11, 3, code],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("checkJoinCode", params, 4)
        return self.postPackDataAndGetUnpackRespData(
            self.LINE_SQUARE_ENDPOINT,
            sqrd,
            4,
            encType=0,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def createSquareChatAnnouncement(
        self,
        squareChatMid: str,
        messageId: str,
        text: str,
        senderSquareMemberMid: str,
        createdAt: int,
        announcementType: int = 0,
    ):
        """
        - SquareChatAnnouncementType:
            TEXT_MESSAGE(0);
        """
        params = [
            [
                12,
                1,
                [
                    [8, 1, self.getCurrReqId()],
                    [11, 2, squareChatMid],
                    [
                        12,
                        3,
                        [
                            [8, 2, announcementType],
                            [
                                12,
                                3,
                                [
                                    [
                                        12,
                                        1,
                                        [
                                            [11, 1, messageId],
                                            [11, 2, text],
                                            [11, 3, senderSquareMemberMid],
                                            [10, 4, createdAt],
                                        ],
                                    ]
                                ],
                            ],
                        ],
                    ],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol(
            "createSquareChatAnnouncement", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareAuthority(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareAuthority is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareAuthority", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareChat", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def refreshSquareSubscriptions(self, subscriptions: list):
        params = [
            [
                12,
                1,
                [
                    [15, 2, [10, subscriptions]],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("refreshSubscriptions", params, 4)
        return self.postPackDataAndGetUnpackRespData(
            self.LINE_SQUARE_ENDPOINT,
            sqrd,
            4,
            encType=0,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getJoinedSquareChats(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getJoinedSquareChats is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getJoinedSquareChats", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def joinSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("joinSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "joinSquareChat", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def findSquareByEmid(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("findSquareByEmid is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "findSquareByEmid", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareMemberRelation(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareMemberRelation is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareMemberRelation", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareMember(self, squareMemberMid: str):
        params = [
            [
                12,
                1,
                [
                    [11, 2, squareMemberMid],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("getSquareMember", params, 4)
        return self.postPackDataAndGetUnpackRespData(
            self.LINE_SQUARE_ENDPOINT,
            sqrd,
            4,
            encType=0,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def destroySquareMessages(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("destroyMessages is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "destroyMessages", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareCategories(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getCategories is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getCategories", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def reportSquareMember(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("reportSquareMember is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "reportSquareMember", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareNoteStatus(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getNoteStatus is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getNoteStatus", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def searchSquareChatMembers(
        self,
        squareChatMid: str,
        displayName: str = "",
        continuationToken: str = None,
        limit: int = 20,
    ):
        SearchSquareChatMembersRequest = [
            [11, 1, squareChatMid],
            [
                12,
                2,
                [
                    [11, 1, displayName],
                    # [2, 2, True]  # includingMe
                ],
            ],
            [8, 4, limit],
        ]
        if continuationToken is not None:
            SearchSquareChatMembersRequest.append([11, 3, continuationToken])
        params = [[12, 1, SearchSquareChatMembersRequest]]
        sqrd = self.generateDummyProtocol(
            "searchSquareChatMembers", params, self.SquareService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.SquareService_API_PATH,
            sqrd,
            self.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
        )

    def getSquareChatFeatureSet(self, squareChatMid: str):
        """Get square chat feature set."""
        METHOD_NAME = "getSquareChatFeatureSet"
        params = [
            [11, 2, squareChatMid],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def getSquareEmid(self, squareMid: str):
        """
        Get square eMid.

        GENERATED BY YinMo0913_DeachSword-DearSakura_v1.0.5.py
        DATETIME: 02/03/2023, 23:02:07
        """
        METHOD_NAME = "getSquareEmid"
        params = [[11, 1, squareMid]]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def getSquareMembersBySquare(self, squareMid: str, squareMemberMids: list):
        """
        Get square members by square.

        GENERATED BY YinMo0913_DeachSword-DearSakura_v1.0.5.py
        DATETIME: 02/03/2023, 23:02:07
        """
        METHOD_NAME = "getSquareMembersBySquare"
        params = [
            [11, 2, squareMid],
            [14, 3, [11, squareMemberMids]],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)

    def manualRepair(self, syncToken: str, limit: int, continuationToken: str):
        """
        Manual repair.

        GENERATED BY YinMo0913_DeachSword-DearSakura_v1.0.5.py
        DATETIME: 02/03/2023, 23:02:07
        """
        METHOD_NAME = "manualRepair"
        params = [
            [11, 1, syncToken],
            [8, 2, limit],
            [11, 3, continuationToken],
        ]
        return SquareServiceStruct.SendRequestByName(self, METHOD_NAME, params)


class SquareServiceStruct(BaseServiceStruct):
    @staticmethod
    def UnsendMessageRequest(squareChatMid: str, messageId: str):
        return __class__.BaseRequest([[11, 2, squareChatMid], [11, 3, messageId]])

    @staticmethod
    def SendRequestByName(client: "CHRLINE", name: str, request: str):
        payload = __class__.BaseRequest(request)
        sqrd = client.generateDummyProtocol(
            name, payload, client.SquareService_REQ_TYPE
        )
        return client.postPackDataAndGetUnpackRespData(
            client.SquareService_API_PATH,
            sqrd,
            client.SquareService_RES_TYPE,
            baseException=SquareService.SQUARE_EXCEPTION,
            readWith=f"SquareService.{name}",
        )



# -*- coding: utf-8 -*-
from .services.TestService import TestService
from .services.TalkService import TalkService
from .services.SquareService import SquareService
from .services.SquareBotService import SquareBotService
from .services.ShopService import ShopService
from .services.SettingsService import SettingsService
from .services.SecondaryPwlessLoginService import SecondaryPwlessLoginService
from .services.SecondaryPwlessLoginPermitNoticeService import (
    SecondaryPwlessLoginPermitNoticeService,
)
from .services.PrimaryAccountInitService import PrimaryAccountInitService
from .services.LiffService import LiffService
from .services.E2EEKeyBackupService import E2EEKeyBackupService
from .services.ChatAppService import ChatAppService
from .services.ChannelService import ChannelService
from .services.CallService import CallService
from .services.BuddyService import BuddyService
from .services.AuthService import AuthService
from .services.AccountAuthFactorEapConnectService import (
    AccountAuthFactorEapConnectService,
)
from .services.AccessTokenRefreshService import AccessTokenRefreshService
from .services.HomeSafetyCheckService import HomeSafetyCheckService
from .services.PrimaryQrCodeMigrationLongPollingService import (
    PrimaryQrCodeMigrationLongPollingService,
)
from .services.PrimaryQrCodeMigrationPreparationService import (
    PrimaryQrCodeMigrationPreparationService,
)
from .services.LoginService import LoginService
from .services.InterlockService import InterlockService
from .services.RelationService import RelationService
from .services.SquareLiveTalkService import SquareLiveTalkService
from .server import Server
from .exceptions import LineServiceException
import rsa
import requests
import httpx
import base64
import binascii


class API(
    TalkService,
    ShopService,
    LiffService,
    ChannelService,
    SquareService,
    BuddyService,
    PrimaryAccountInitService,
    AuthService,
    SettingsService,
    AccessTokenRefreshService,
    CallService,
    SecondaryPwlessLoginService,
    SecondaryPwlessLoginPermitNoticeService,
    ChatAppService,
    AccountAuthFactorEapConnectService,
    E2EEKeyBackupService,
    SquareBotService,
    TestService,
    HomeSafetyCheckService,
    PrimaryQrCodeMigrationLongPollingService,
    PrimaryQrCodeMigrationPreparationService,
    LoginService,
    InterlockService,
    RelationService,
    SquareLiveTalkService
):
    _msgSeq = 0
    url = "https://gf.line.naver.jp/enc"

    def __init__(self, forwardedIp=None, proxy_string=None):                    
        self.server = Server()
        if proxy_string is None:
            self.req = requests.session()
        else:
            self.req = requests.session()
            self.req.proxies = {
                "http": proxy_string,
                "https": proxy_string
            }
            
        self.req_h2 = httpx.Client(http2=True)
        self.server.Headers = {
            "x-line-application": self.APP_NAME,
            "x-le": self.le,
            "x-lap": "5",
            # "x-lc": "3",
            # "X-LST": "",
            # "X-LCR": "",
            # "X-LOR": "",
            "x-lpv": "1",
            "x-lcs": self._encryptKey,
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
            "content-type": "application/x-thrift; protocol=TBINARY",
            "x-lal": self.LINE_LANGUAGE,
            "x-lhm": "POST",
        }
        if forwardedIp is not None:
            self.server.Headers["X-Forwarded-For"] = forwardedIp
        self.authToken = None
        self.revision = 0
        self.globalRev = 0
        self.individualRev = 0
        self._msgSeq = 0
        TalkService.__init__(self)
        ShopService.__init__(self)
        LiffService.__init__(self)
        ChannelService.__init__(self)
        SquareService.__init__(self)
        BuddyService.__init__(self)
        PrimaryAccountInitService.__init__(self)
        AuthService.__init__(self)
        SettingsService.__init__(self)
        AccessTokenRefreshService.__init__(self)
        CallService.__init__(self)
        SecondaryPwlessLoginService.__init__(self)
        SecondaryPwlessLoginPermitNoticeService.__init__(self)
        ChatAppService.__init__(self)
        AccountAuthFactorEapConnectService.__init__(self)
        E2EEKeyBackupService.__init__(self)
        SquareBotService.__init__(self)
        TestService.__init__(self)
        HomeSafetyCheckService.__init__(self)
        PrimaryQrCodeMigrationLongPollingService.__init__(self)
        PrimaryQrCodeMigrationPreparationService.__init__(self)
        LoginService.__init__(self)
        InterlockService.__init__(self)
        RelationService.__init__(self)
        SquareLiveTalkService.__init__(self)

    def requestPwlessLogin(self, phone, pw):
        pwless_code = self.checkAndGetValue(self.createPwlessSession(phone), 1, "val_1")
        print(f"PWLESS SESSION: {pwless_code}")
        cert = self.getCacheData(".pwless", phone)
        try:
            self.verifyLoginCertificate(pwless_code, "" if cert is None else cert)
        except:
            pwless_pincode = self.checkAndGetValue(
                self.requestPinCodeVerif(pwless_code), 1, "val_1"
            )
            print(f"PWLESS PINCODE: {pwless_pincode}")
            self.checkPwlessPinCodeVerified(pwless_code)
        secret, secretPK = self.createSqrSecret(True)
        self.putExchangeKey(pwless_code, secretPK)
        self.requestPaakAuth(pwless_code)
        print(f"need Paak Auth Confind")
        self.checkPaakAuthenticated(pwless_code)
        ek = self.getE2eeKey(pwless_code)
        try:
            loginInfo = self.pwlessLoginV2(pwless_code)
            cert = self.checkAndGetValue(loginInfo, 2, "val_2")
            tokenInfo = self.checkAndGetValue(loginInfo, 3, "val_3")
            token = self.checkAndGetValue(tokenInfo, 1, "val_1")
            token2 = self.checkAndGetValue(tokenInfo, 2, "val_2")
            mid = self.checkAndGetValue(loginInfo, 5, "val_5")
        except:
            loginInfo = self.pwlessLogin(pwless_code)
            token = self.checkAndGetValue(loginInfo, 1, "val_1")
            cert = self.checkAndGetValue(loginInfo, 2, "val_2")
            mid = self.checkAndGetValue(loginInfo, 4, "val_4")
        self.authToken = token
        print(f"Auth Token: {self.authToken}")
        self.saveCacheData(".pwless", phone, cert)
        self.decodeE2EEKeyV1(self.checkAndGetValue(ek, 1, "val_1"), secret, mid)
        return True

    def requestEmailLogin(self, email, pw, e2ee=True):
        rsaKey = self.getRSAKeyInfo()
        keynm = self.checkAndGetValue(rsaKey, 1, "val_1")
        nvalue = self.checkAndGetValue(rsaKey, 2, "val_2")
        evalue = self.checkAndGetValue(rsaKey, 3, "val_3")
        sessionKey = self.checkAndGetValue(rsaKey, 4, "val_4")
        certificate = self.getEmailCert(email)
        message = (
            chr(len(sessionKey))
            + sessionKey
            + chr(len(email))
            + email
            + chr(len(pw))
            + pw
        ).encode("utf-8")
        pub_key = rsa.PublicKey(int(nvalue, 16), int(evalue, 16))
        crypto = binascii.hexlify(rsa.encrypt(message, pub_key)).decode()
        secret, secretPK = self.createSqrSecret(True)
        pincode = b"202202"
        _secret = self._encryptAESECB(
            self.getSHA256Sum(pincode), base64.b64decode(secretPK)
        )
        _req = {
            "keynm": keynm,
            "encData": crypto,
            "secret": _secret,
            "deviceName": self.SYSTEM_NAME,
            "calledName": "loginZ",
        }
        if not e2ee:
            _req["secret"] = None
        try:
            res = self.loginV2(**_req, cert=certificate)
        except LineServiceException as e:
            res = {}
            if e.code == 89:
                if not e2ee:
                    raise e
                return self.requestEmailLogin(email, pw, False)
            raise e
        if self.checkAndGetValue(res, 1, "val_1") is None:
            verifier = self.checkAndGetValue(res, 3, "val_3")
            if not e2ee:
                pincode = self.checkAndGetValue(res, 4).encode()
            print(f"Enter Pincode: {pincode.decode()}")
            if e2ee:
                e2eeInfo = self.checkLoginV2PinCode(verifier)["metadata"]
                try:
                    self.decodeE2EEKeyV1(e2eeInfo, secret)
                except:
                    raise Exception(f"e2eeInfo decode failed, try again")
                blablabao = self.encryptDeviceSecret(
                    base64.b64decode(e2eeInfo["publicKey"]),
                    secret,
                    base64.b64decode(e2eeInfo["encryptedKeyChain"]),
                )
                e2eeLogin = self.confirmE2EELogin(verifier, blablabao)
            else:
                e2eeLogin = self.checkLoginZPinCode(verifier)["verifier"]
            try:
                res = self.loginV2(**_req, verifier=e2eeLogin)
                self.saveEmailCert(email, self.checkAndGetValue(res, 2, "val_2"))
            except:
                raise Exception(f"confirmE2EELogin failed, try again")
        self.authToken = self.checkAndGetValue(res, 1)
        print(f"AuthToken: {self.authToken}")
        return True

    def requestEmailLoginV2(self, email, pw):
        rsaKey = self.getRSAKeyInfo()
        keynm = self.checkAndGetValue(rsaKey, 1, "val_1")
        nvalue = self.checkAndGetValue(rsaKey, 2, "val_2")
        evalue = self.checkAndGetValue(rsaKey, 3, "val_3")
        sessionKey = self.checkAndGetValue(rsaKey, 4, "val_4")
        certificate = self.getEmailCert(email)
        message = (
            chr(len(sessionKey))
            + sessionKey
            + chr(len(email))
            + email
            + chr(len(pw))
            + pw
        ).encode("utf-8")
        pub_key = rsa.PublicKey(int(nvalue, 16), int(evalue, 16))
        crypto = binascii.hexlify(rsa.encrypt(message, pub_key)).decode()
        secret, secretPK = self.createSqrSecret(True)
        pincode = b"1314520"
        _secret = self._encryptAESECB(
            self.getSHA256Sum(pincode), base64.b64decode(secretPK)
        )
        try:
            res = self.loginV2(
                keynm, crypto, _secret, deviceName=self.SYSTEM_NAME, cert=certificate
            )
        except LineServiceException as e:
            if e.code == 89:
                print(f"can't login: {e.message}, try use LoginZ...")
                return self.requestEmailLogin(email, pw)
            raise e
        if self.checkAndGetValue(res, 9, "val_9") is None:
            verifier = self.checkAndGetValue(res, 3, "val_3")
            if self.checkAndGetValue(res, 5, "val_5") == 3:
                print(f"need device confirm")
            print(f"Enter Pincode: {pincode.decode()}")
            e2eeInfo = self.checkLoginV2PinCode(verifier)["metadata"]
            try:
                e2eeKeyInfo = self.decodeE2EEKeyV1(e2eeInfo, secret)
            except:
                raise Exception(f"e2eeInfo decode failed, try again")
            blablabao = self.encryptDeviceSecret(
                base64.b64decode(e2eeInfo["publicKey"]),
                secret,
                base64.b64decode(e2eeInfo["encryptedKeyChain"]),
            )
            e2eeLogin = self.confirmE2EELogin(verifier, blablabao)
            try:
                res = self.loginV2(
                    None, None, None, deviceName=self.SYSTEM_NAME, verifier=e2eeLogin
                )
            except LineServiceException as e:
                print(e)
                if e.code in [20, 89]:
                    print(f"can't login: {e.message}, try use LoginZ...")
                    return self.requestEmailLogin(email, pw)
            self.saveEmailCert(email, self.checkAndGetValue(res, 2, "val_2"))
        loginInfo = self.checkAndGetValue(res, 9, "val_9")
        self.authToken = self.checkAndGetValue(loginInfo, 1, "val_1")
        refreshToken = self.checkAndGetValue(loginInfo, 2, "val_2")
        self.saveCacheData(".refreshToken", self.authToken, refreshToken)
        print(f"AuthToken: {self.authToken}")
        print(f"RefreshToken: {refreshToken}")
        return True

    def requestSQR(self, isSelf=True):
        sqr = self.checkAndGetValue(self.createSession(), 1, "val_1")
        url = self.checkAndGetValue(self.createQrCode(sqr), 1, "val_1")
        secret, secretUrl = self.createSqrSecret()
        url = url + secretUrl
        imgPath = self.genQrcodeImageAndPrint(url)
        yield f"URL: {url}"
        yield f"IMG: {imgPath}"
        if self.checkQrCodeVerified(sqr):
            try:
                self.verifyCertificate(sqr, self.getSqrCert())
            except:
                c = self.checkAndGetValue(self.createPinCode(sqr), 1, "val_1")
                yield f"請輸入pincode: {c}"
                self.checkPinCodeVerified(sqr)
            e = self.qrCodeLogin(sqr, secret)
            if isSelf:
                self.authToken = e
                print(f"AuthToken: {self.authToken}")
            else:
                yield e
            return
            raise Exception("can not check pin code, try again?")
        raise Exception("can not check qr code, try again?")

    def requestSQR2(self, isSelf=True):
        sqr = self.checkAndGetValue(self.createSession(), 1, "val_1")
        url = self.checkAndGetValue(self.createQrCode(sqr), 1, "val_1")
        secret, secretUrl = self.createSqrSecret()
        url = url + secretUrl
        imgPath = self.genQrcodeImageAndPrint(url)
        yield f"URL: {url}"
        yield f"IMG: {imgPath}"
        if self.checkQrCodeVerified(sqr):
            try:
                self.verifyCertificate(sqr, self.getSqrCert())
            except:
                c = self.createPinCode(sqr)
                yield f"請輸入pincode: {c}"
                self.checkPinCodeVerified(sqr)
            try:
                e = self.qrCodeLoginV2(sqr, self.APP_TYPE, self.SYSTEM_NAME, True)
                cert = self.checkAndGetValue(e, 1)
                self.saveSqrCert(cert)
                tokenV3Info = self.checkAndGetValue(e, 3)
                _mid = self.checkAndGetValue(e, 4)
                bT = self.checkAndGetValue(e, 9)
                metadata = self.checkAndGetValue(e, 10)
                e2eeKeyInfo = self.decodeE2EEKeyV1(metadata, secret)
                authToken = self.checkAndGetValue(tokenV3Info, 1)
                refreshToken = self.checkAndGetValue(tokenV3Info, 2)
                self.saveCacheData(".refreshToken", authToken, refreshToken)
                print(f"AuthToken: {authToken}")
                print(f"RefreshToken: {refreshToken}")
                if isSelf:
                    self.authToken = authToken
                yield authToken
            except LineServiceException as e:
                print(e)
                yield "try using requestSQR()..."
                for _ in self.requestSQR(isSelf):
                    yield _
            return
            raise Exception("can not check pin code, try again?")
        raise Exception("can not check qr code, try again?")

    def requestSQR3(self, isSelf=True):
        """
        Request Secondary QrCode Login for secure.
        
        Source: https://github.com/DeachSword/CHRLINE/blob/445e433b1fbe9a020f6bc1cbd0eb7af3f75ce196/examples/test_sqr_4_secure.py
        """
        log4Debug = False
        sqr = self.checkAndGetValue(self.createSession(), 1)
        qr4s = self.createQrCodeForSecure(sqr)
        self.log(qr4s, log4Debug)
        url = self.checkAndGetValue(qr4s, 'callbackUrl', 1)
        nonce = self.checkAndGetValue(qr4s, 'nonce', 4)
        self.log(f"nonce: {nonce}", log4Debug)
        secret, secretUrl = self.createSqrSecret()
        url = url + secretUrl
        imgPath = self.genQrcodeImageAndPrint(url)
        yield f"URL: {url}"
        yield f"IMG: {imgPath}"
        if self.checkQrCodeVerified(sqr):
            try:
                self.verifyCertificate(sqr, self.getSqrCert())
            except:
                c = self.createPinCode(sqr)
                yield f"請輸入pincode: {c}"
                self.checkPinCodeVerified(sqr)
            e = self.qrCodeLoginV2ForSecure(sqr, self.APP_TYPE, self.SYSTEM_NAME, nonce)
            self.log(e, log4Debug)
            cert = self.checkAndGetValue(e, 'certificate', 1)
            self.saveSqrCert(cert)
            tokenV3Info = self.checkAndGetValue(e, 'tokenV3IssueResult', 3)
            _mid = self.checkAndGetValue(e, 'mid', 4)
            metadata = self.checkAndGetValue(e, 'metaData', 10)
            self.decodeE2EEKeyV1(metadata, secret, _mid)
            authToken = self.checkAndGetValue(tokenV3Info, 'accessToken', 1)
            refreshToken = self.checkAndGetValue(tokenV3Info, 'refreshToken', 2)
            self.saveCacheData(".refreshToken", authToken, refreshToken)
            print(f"AuthToken: {authToken}")
            print(f"RefreshToken: {refreshToken}")
            if isSelf:
                self.authToken = authToken
            yield authToken
            return
        raise Exception("can not check qr code, try again?")

    def createSession(self):
        params = []
        sqrd = self.generateDummyProtocol("createSession", params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/lgn/sq/v1", sqrd, 3)

    def createQrCode(self, qrcode):
        params = [[12, 1, [[11, 1, qrcode]]]]
        sqrd = self.generateDummyProtocol("createQrCode", params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/lgn/sq/v1", sqrd, 3)

    def checkQrCodeVerified(self, qrcode):
        params = [
            [
                12,
                1,
                [
                    [11, 1, qrcode],
                ],
            ],
        ]
        sqrd = self.generateDummyProtocol("checkQrCodeVerified", params, 3)
        headers = self.server.additionalHeaders(
            self.server.Headers, {"x-lst": "150000"}  # timeout
        )
        try:
            self.postPackDataAndGetUnpackRespData(
                "/acct/lp/lgn/sq/v1",
                sqrd,
                3,
                headers=headers,
                access_token=qrcode,
                conn=requests.session(),
            )
        except Exception as e:
            print(f"[checkQrCodeVerified] {e}")
            return False
        return True

    def verifyCertificate(self, qrcode, cert=None):
        params = [
            [
                12,
                1,
                [
                    [11, 1, qrcode],
                    [11, 2, cert],
                ],
            ],
        ]
        sqrd = self.generateDummyProtocol("verifyCertificate", params, 3)
        return self.postPackDataAndGetUnpackRespData(
            self.LINE_SECONDARY_QR_LOGIN_ENDPOINT, sqrd, 3
        )

    def createPinCode(self, qrcode):
        params = [
            [
                12,
                1,
                [
                    [11, 1, qrcode],
                ],
            ],
        ]
        sqrd = self.generateDummyProtocol("createPinCode", params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/lgn/sq/v1", sqrd, 3)

    def checkPinCodeVerified(self, qrcode):
        params = [
            [
                12,
                1,
                [
                    [11, 1, qrcode],
                ],
            ],
        ]
        sqrd = self.generateDummyProtocol("checkPinCodeVerified", params, 3)
        headers = self.server.additionalHeaders(
            self.server.Headers, {"x-lst": "150000"}  # timeout
        )
        try:
            self.postPackDataAndGetUnpackRespData(
                "/acct/lp/lgn/sq/v1",
                sqrd,
                3,
                headers=headers,
                access_token=qrcode,
                conn=requests.session(),
            )
        except Exception as e:
            print(f"[checkPinCodeVerified] {e}")
            return False
        return True

    def qrCodeLogin(
        self, authSessionId: str, secret: str, autoLoginIsRequired: bool = True
    ):

        params = [
            [
                12,
                1,
                [
                    [11, 1, authSessionId],
                    [11, 2, self.SYSTEM_NAME],
                    [2, 3, autoLoginIsRequired],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("qrCodeLogin", params, 3)
        data = self.postPackDataAndGetUnpackRespData("/acct/lgn/sq/v1", sqrd, 3)
        pem = self.checkAndGetValue(data, 1, "val_1")
        self.saveSqrCert(pem)
        print("證書: ", pem)
        _token = self.checkAndGetValue(data, 2, "val_2")
        e2eeInfo = self.checkAndGetValue(data, 4, "val_4")
        _mid = self.checkAndGetValue(data, 5, "val_5")
        if e2eeInfo is not None:
            self.decodeE2EEKeyV1(e2eeInfo, secret, _mid)
        return _token

    def qrCodeLoginV2(
        self,
        authSessionId,
        modelName="彥彥好睡",
        systemName="鴻鴻好暈",
        autoLoginIsRequired=True,
    ):
        params = [
            [
                12,
                1,
                [
                    [11, 1, authSessionId],
                    [11, 2, systemName],
                    [11, 3, modelName],
                    [2, 4, autoLoginIsRequired],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol("qrCodeLoginV2", params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/lgn/sq/v1", sqrd, 3)

    def CPF(self):
        sqrd = []
        return self.postPackDataAndGetUnpackRespData("/CPF", sqrd)

    def getRSAKeyInfo(self, provider=1):
        """
        provider:
         - UNKNOWN(0),
         - LINE(1),
         - NAVER_KR(2),
         - LINE_PHONE(3)
        """
        params = [
            [8, 2, provider],
        ]
        sqrd = self.generateDummyProtocol("getRSAKeyInfo", params, 3)
        return self.postPackDataAndGetUnpackRespData("/api/v3/TalkService.do", sqrd, 3)

    def loginV2(
        self,
        keynm,
        encData,
        secret,
        deviceName="Chrome",
        cert=None,
        verifier=None,
        calledName="loginV2",
    ):
        loginType = 2
        if secret is None:
            loginType = 0
        if verifier is not None:
            loginType = 1
        params = [
            [
                12,
                2,
                [
                    [8, 1, loginType],
                    [8, 2, 1],  # provider
                    [11, 3, keynm],
                    [11, 4, encData],
                    [2, 5, 0],
                    [11, 6, ""],
                    [11, 7, deviceName],
                    [11, 8, cert],
                    [11, 9, verifier],
                    [11, 10, secret],
                    [8, 11, 1],
                    [11, 12, "System Product Name"],
                ],
            ]
        ]
        sqrd = self.generateDummyProtocol(calledName, params, 3)
        return self.postPackDataAndGetUnpackRespData("/api/v3p/rs", sqrd, 3)

    def loginZ(
        self,
        keynm,
        encData,
        systemName="DeachSword-2021",
        certificate=None,
        verifier=None,
    ):
        sqrd = [128, 1, 0, 1] + self.getStringBytes("loginZ") + [0, 0, 0, 0]
        sqrd += [12, 0, 2]
        loginType = 0
        if verifier is not None:
            loginType = 1
        sqrd += [8, 0, 1] + self.getIntBytes(loginType)  # 2 if e2ee
        sqrd += [8, 0, 2] + self.getIntBytes(1)  # provider
        sqrd += [11, 0, 3] + self.getStringBytes(keynm)
        sqrd += [11, 0, 4] + self.getStringBytes(encData)
        sqrd += [2, 0, 5, 0]
        sqrd += [11, 0, 6] + self.getStringBytes("")  # accessLocation
        sqrd += [11, 0, 7] + self.getStringBytes(systemName)
        sqrd += [11, 0, 8] + self.getStringBytes(certificate)
        if verifier is not None:
            sqrd += [11, 0, 9] + self.getStringBytes(verifier)
        # sqrd += [11, 0, 10] + self.getStringBytes("") #secret
        sqrd += [8, 0, 11] + self.getIntBytes(1)
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData("/api/v3p/rs", sqrd, 3)

    def checkLoginZPinCode(self, accessSession):
        hr = self.server.additionalHeaders(self.server.Headers, {"x-lhm": "GET"})
        return self.postPackDataAndGetUnpackRespData(
            self.SECONDARY_DEVICE_LOGIN_VERIFY_PIN,
            [],
            -3,
            headers=hr,
            access_token=accessSession,
            conn=requests.session(),
        )["result"]

    def checkLoginV2PinCode(self, accessSession):
        hr = self.server.additionalHeaders(self.server.Headers, {"x-lhm": "GET"})
        return self.postPackDataAndGetUnpackRespData(
            self.SECONDARY_DEVICE_LOGIN_VERIFY_PIN_WITH_E2EE,
            [],
            -3,
            headers=hr,
            access_token=accessSession,
            conn=requests.session(),
        )["result"]

    def testTBinary(self):
        METHOD_NAME = "getProfile"
        params = []
        sqrd = self.generateDummyProtocol(METHOD_NAME, params, 3)
        res = self.postPackDataAndGetUnpackRespData("/S3", sqrd, -1)
        return res

    def testTCompact(self):
        METHOD_NAME = "getProfile"
        params = []
        sqrd = self.generateDummyProtocol(METHOD_NAME, params, 4)
        res = self.postPackDataAndGetUnpackRespData("/S4", sqrd, -1)
        return res

    def testTMoreCompact(self):
        METHOD_NAME = "getProfile"
        params = []
        sqrd = self.generateDummyProtocol(METHOD_NAME, params, 4)
        res = self.postPackDataAndGetUnpackRespData("/S5", sqrd, -1)
        return res



# -*- coding: utf-8 -*-
import json
import time
import hashlib
from base64 import b64encode
import uuid
import urllib
import httpx


class Object(object):
    def __init__(self):
        self.Hraders4Obs = {
            "User-Agent": self.server.Headers["User-Agent"],
            "X-Line-Access": self.authToken,
            "X-Line-Application": self.server.Headers["x-line-application"],
            "X-Line-Mid": self.mid,
            "x-lal": self.LINE_LANGUAGE,
        }
        if self.DEVICE_TYPE != "CHROMEOS":
            self.Hraders4Obs[
                "X-Line-Access"
            ] = self.acquireEncryptedAccessToken().split("\x1e")[1]
        self.obsConn = httpx.Client(http2=True, timeout=None)

    """ 

    TimelineObs 
    Source: https://github.com/DeachSword/LINE-DemoS-Bot/blob/master/api/obs/obs.py

    """

    def sendImage(self, to, path, oType="image"):
        return self.uploadObjTalk(pathOrBytes=path, oType=oType, to=to)

    def sendGIF(self, to, path, oType="gif"):
        return self.uploadObjTalk(pathOrBytes=path, oType=oType, to=to)

    def sendVideo(self, to, path, oType="video"):
        return self.uploadObjTalk(pathOrBytes=path, oType=oType, to=to)

    def sendAudio(self, to, path, oType="audio"):
        return self.uploadObjTalk(pathOrBytes=path, oType=oType, to=to)

    def sendFile(self, to, path, oType="file"):
        return self.uploadObjTalk(pathOrBytes=path, oType=oType, to=to)

    def sendObjHomeToTalk(self, to, oid, returnHeaders=False, **kwargs):
        data = {
            "oid": "reqseq",
            "tomid": to,
            "reqseq": self.getCurrReqId(),
            "ver": "1.0",
        }
        if (
            "issueToken4ChannelId" not in kwargs
        ):  # If you don't want this value, please set it to None.
            kwargs["issueToken4ChannelId"] = "1341209850"
        kwargs["copyFromService"] = "myhome"
        kwargs["copyFromSid"] = "h"
        kwargs["copyFromObjId"] = oid
        kwargs["copy2Service"] = "talk"
        kwargs["copy2Sid"] = "m"
        kwargs["copy2ObjId"] = "reqseq"
        obsObjId, obsHash, respHeaders = self.forwardObjectForService(data, **kwargs)
        if returnHeaders:
            return respHeaders
        return obsObjId

    def updateProfileImage(self, path, storyShare=False, type="p", mid: str = None):
        """
        Update profile image.

        use `mid` for square member.
        """
        if mid is None:
            mid = self.mid
        midType = self.getToType(mid)
        if midType == 0:
            obsPath = f"talk/p/{mid}"
        elif midType == 5:
            obsPath = f"g2/member/{mid}"
        else:
            raise ValueError(f"Invalid midType: {midType}")
        oType = "IMAGE"
        params = {}
        is_video = self.checkIsVideo(path)
        hstr = "DearSakura_%s" % int(time.time() * 1000)
        hstr = hashlib.md5(hstr.encode()).hexdigest()
        filename = "profile.jpg"
        metaData = {"profileContext": {"storyShare": storyShare}}
        if type == "vp":
            params["cat"] = "vp.mp4"  # let server know u have vp.mp4
        is_video = self.checkIsVideo(path)
        if is_video:
            obsPath = f"talk/vp/{mid}"
            oType = "VIDEO"
            filename = f"{hstr}.mp4"
            params = {
                "cat": "vp.mp4",
            }
        talkMeta = b64encode(json.dumps(metaData).encode("utf-8")).decode("utf-8")
        try:
            objId, objHash, respHeaders = self.uploadObjectForService(
                path, oType, obsPath, "1341209850", params, talkMeta, filename
            )
        except Exception as e:
            raise e
        return True

    def updateProfileCover(self, path, storyShare=False):
        hstr = "DearSakura_%s" % int(time.time() * 1000)
        objId = hashlib.md5(hstr.encode()).hexdigest()
        obsPath = f"myhome/c/{objId}"
        oType = "IMAGE"
        filename = f"{objId}.jpg"
        params = None
        is_video = self.checkIsVideo(path)
        if is_video:
            obsPath = f"myhome/vc/{objId}"
            oType = "VIDEO"
            filename = f"{objId}.mp4"
            params = {}
        try:
            objId, objHash, respHeaders = self.uploadObjectForService(
                path, oType, obsPath, "1341209850", params, filename=filename
            )
            print(respHeaders)
        except Exception as e:
            raise Exception(e)
        if is_video:
            _url, _vc_url, _objId, _vc_objId = self.getProfileCoverObjIdAndUrl(self.mid)
            home = self.updateProfileCoverById(
                _objId, f"{objId}/vp", storyShare=storyShare
            )
        else:
            home = self.updateProfileCoverById(objId, storyShare=storyShare)
        return home

    def updateChatProfileImage(self, groupId: str, path: str, oType: str = "IMAGE"):
        midType = self.getToType(groupId)
        if midType in [1, 2]:
            obsPath = f"talk/g/{groupId}"
        elif midType == 3:
            obsPath = f"g2/group/{groupId}"
        elif midType == 4:
            obsPath = f"g2/chat/{groupId}"
        else:
            raise ValueError(f"Invalid midType: {midType}")
        params = {"cat": "original"}
        try:
            objId, objHash, respHeaders = self.uploadObjectForService(
                path, oType, obsPath, "1341209850", params, filename="profile.jpg"
            )
        except Exception as e:
            raise Exception(f"updateChatProfileImage failure: {e}")
        return True

    def updateImage2Album(
        self,
        homeId: str,
        albumId: int,
        pathOrBytes: any,
        oType: str = "IMAGE",
        updateAlbum: bool = True,
    ):
        """
        Upload Object to Album Obs.

        will use `updateAlbum` to update the album, if content uploads success.
        ---------------
        Return `OBS_OID` and `OBS_HASH`
        """
        oids = []
        additionalHeaders = {
            "X-Line-Album": str(albumId),
            "X-Line-Mid": homeId,
        }
        if type(pathOrBytes) != list:
            pathOrBytes = [pathOrBytes]
        for _uploadData in pathOrBytes:
            oid = f"{uuid.uuid1().hex}.20{time.strftime('%m%d')}08"
            params = {"oid": oid}
            obsObjId, obsHash, respHeaders = self.uploadObjectForService(
                _uploadData,
                oType,
                f"album/a/{oid}",
                issueToken4ChannelId="1341209850",
                params=params,
                additionalHeaders=additionalHeaders,
            )
            oids.append(obsObjId)
        if updateAlbum:
            self.addImageToAlbum(homeId, albumId, oids)
        return oids

    def uploadObjHome(self, path, type="image", objId=None):
        if type == "image":
            contentType = "image/jpeg"
        elif type == "video":
            contentType = "video/mp4"
        elif type == "audio":
            contentType = "audio/mp3"
        else:
            raise Exception("Invalid type value")
        if not objId:
            hstr = "DearSakura_%s" % int(time.time() * 1000)
            objId = hashlib.md5(hstr.encode()).hexdigest()
        file = open(path, "rb").read()
        params = {
            "name": f"{objId}.mp4",
            "oid": "%s" % str(objId),
            "type": type,
            "ver": "2.0",  # 2.0 :p
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "Content-Type": contentType,
                "Content-Length": str(len(file)),
                "x-obs-params": self.genOBSParams(params, "b64"),  # base64 encode
            },
        )
        r = self.server.postContent(
            self.LINE_OBS_DOMAIN + "/myhome/h/upload.nhn", headers=hr, data=file
        )
        if r.status_code != 201:
            raise Exception(
                f"Upload object home failure. Receive statue code: {r.status_code}"
            )
        return objId

    def uploadMultipleImageToTalk(
        self, pathOrObjids: list, to: str, mtype: str = "image", oTypes: list = None
    ):
        """
        Args:
            pathOrObjids: multiple image paths or multiple objids according to the type
            to: send chat id
            mtype: image or objids
            oTypes: image
        """

        def createTalkMeta(hmap):
            sqrd_base = [13, 0, 18, 11, 11]
            sqrd = sqrd_base + self.getIntBytes(len(hmap.keys()))
            for hm in hmap.keys():
                sqrd += self.getStringBytes(hm)
                sqrd += self.getStringBytes(hmap[hm])
            sqrd += [0]
            data = bytes(sqrd)
            msg = json.dumps({"message": b64encode(data).decode("utf-8")})
            return b64encode(msg.encode("utf-8")).decode("utf-8")

        if mtype.lower() == "image":
            mtype = "image"
        elif mtype.lower() == "objids":
            mtype = "objids"
        else:
            raise Exception("Type must be image or object")

        if type(pathOrObjids) != list:
            raise Exception("pathOrObjids must be list")

        res = {}
        oType = []
        hashmap = {"GID": "0", "GSEQ": "1", "GTOTAL": str(len(pathOrObjids))}
        talkMeta = createTalkMeta(hashmap)
        if mtype == "image":
            if oTypes is None:
                oTypes = ["IMAGE" for _ in range(len(pathOrObjids))]
            oType = iter(oTypes)
            res = self.uploadObjTalk(
                pathOrObjids[0],
                next(oType),
                to=to,
                talkMeta=talkMeta,
                returnHeaders=True,
            )
        elif mtype == "objids":
            res = self.sendObjToTalk(
                to, pathOrObjids[0], talkMeta=talkMeta, returnHeaders=True
            )

        gid, msgIds = res["x-line-message-gid"], [res["x-obs-oid"]]

        if len(pathOrObjids) > 1:
            hashmap["GID"] = gid
            nc = 2
            for poo in pathOrObjids[1:]:
                self.log(f"Upload image-{nc} with GID-{gid}...", True)
                hashmap["GSEQ"] = str(nc)
                talkMeta = createTalkMeta(hashmap)
                if mtype == "image":
                    res = self.uploadObjTalk(
                        poo, next(oType), to=to, talkMeta=talkMeta, returnHeaders=True
                    )
                elif mtype == "objids":
                    res = self.sendObjToTalk(
                        to, poo, talkMeta=talkMeta, returnHeaders=True
                    )
                msgIds.append(res["x-obs-oid"])
                nc += 1

        return gid, msgIds

    def uploadObjTalk(
        self,
        pathOrBytes,
        oType="image",
        objId=None,
        to=None,
        talkMeta=None,
        returnHeaders=False,
        filename: str = None,
        isOriginal: bool = False,
    ):
        if oType.lower() not in ["image", "gif", "video", "audio", "file"]:
            raise Exception("Invalid type value")
        # url = self.LINE_OBS_DOMAIN + '/talk/m/upload.nhn' #if reqseq not working
        # url = self.LINE_OBS_DOMAIN + '/r/talk/m/reqseq'
        serviceName = "talk"
        obsNamespace = "m"
        params = {}
        if oType.lower() == "gif":
            params["cat"] = "original"
            oType = "IMAGE"
        if isOriginal:
            params["cat"] = "original"
        if to is not None:
            params.update(
                {"oid": "reqseq", "reqseq": str(self.getCurrReqId()), "tomid": to}
            )
            objId = "reqseq"
            if self.getToType(to) == 4:
                serviceName = "g2"
        else:
            if objId is not None:
                params["oid"] = objId
        obsObjId, obsHash, respHeaders = self.uploadObjectForService(
            pathOrBytes,
            oType,
            f"{serviceName}/{obsNamespace}/{objId}",
            params=params,
            talkMeta=talkMeta,
            filename=filename,
        )
        if returnHeaders:
            return respHeaders
        objId = obsObjId
        objHash = obsHash  # for view on cdn
        return objId

    def uploadStoryObject(self, pathOrBytes: any, oType="IMAGE"):
        """
        Upload Object to Story Obs.

        ---------------
        Return `OBS_OID` and `OBS_HASH`
        """
        oid = f"{uuid.uuid1().hex}tffffffff"
        params = {"oid": oid}
        obsObjId, obsHash, respHeaders = self.uploadObjectForService(
            pathOrBytes,
            oType,
            f"story/st/{oid}",
            issueToken4ChannelId="1341209850",
            params=params,
        )
        return obsObjId, obsHash

    def forwardObjectMsg(self, to: str, msgId: str, _from: str = "u", **kwargs):
        """
        Forward object to Talk

        use `_from` to specify the source mid of the messageid
        """
        copyFromService = "talk"
        copyFromSid = "m"
        copy2Service = "talk"
        copy2Sid = "m"
        data = {
            "oid": "reqseq",
            "tomid": to,
            "reqseq": self.getCurrReqId(),
            "ver": "1.0",
        }
        toType = self.getToType(to)
        if toType in [0, 1, 2]:
            pass
        elif toType == 4:
            copy2Service = "g2"
        else:
            raise ValueError(f"Invalid `to` midType: {toType}")
        toType = self.getToType(_from)
        if toType in [0, 1, 2]:
            pass
        elif toType == 4:
            copyFromService = "g2"
        else:
            raise ValueError(f"Invalid `_from` midType: {toType}")
        kwargs["copyFromService"] = copyFromService
        kwargs["copyFromSid"] = copyFromSid
        kwargs["copyFromObjId"] = msgId
        kwargs["copy2Service"] = copy2Service
        kwargs["copy2Sid"] = copy2Sid
        kwargs["copy2ObjId"] = "reqseq"
        return self.forwardObjectForService(data, **kwargs)

    def forwardKeepObjectMsg(self, to, oid, contentId, contentType="image"):
        # SHARE KEEP CONTENTS V1
        data = {
            "copyFrom": f"/keep/p/{oid}",
            "tomid": to,
            "reqseq": self.getCurrReqId(),
            "type": contentType,
            "contentId": contentId,
            "size": "-1",
        }
        if contentType == "image":
            data["cat"] = "original"
        elif contentType in ["audio", "video"]:
            data["duration"] = "3000"
        print(self.server.timelineHeaders)
        r = self.server.postContent(
            self.LINE_HOST_DOMAIN + "/oa/r/talk/m/reqseq/copy.nhn",
            data=data,
            headers=self.server.timelineHeaders,
        )
        if not self.checkRespIsSuccessWithLpv(r):
            raise Exception(f"Forward object failure: {r.status_code}")
        return r.headers["X-Obs-Oid"]

    def forwardKeepContent2Mid(self, to: str, oid: str, contentId: str, **kwargs):
        # SHARE KEEP CONTENTS V2
        data = {
            "tomid": to,
            "reqseq": self.getCurrReqId(),
        }
        kwargs["copyFromService"] = "keep"
        kwargs["copyFromSid"] = "p"
        kwargs["copyFromObjId"] = oid
        kwargs["contentId"] = contentId
        return self.forwardObjectForService(data, **kwargs)

    def trainingImage(self, msgId):
        data = {
            "useAsTrainingData": False,
            "input": "obs://talk/m/%s" % msgId,
            "filters": {"OCR": {"params": {"useColorPicker": "true"}}},
        }

        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-client-channel": "chat_viewer",
                "x-lal": "th_TH",
                "x-recognition-input-type": "obs",
                "Content-Type": "application/json",
            },
        )
        r = self.server.postContent(
            self.LINE_OBS_DOMAIN + "/px/talk/m/%s/recognition.obs" % msgId,
            data=json.dumps(data),
            headers=hr,
        )
        if not self.checkRespIsSuccessWithLpv(r):
            raise Exception(f"Training image failure: {r.status_code}")
        return r.json()

    def downloadObjectMsg(self, objId: str, path: str = None, objFrom: str = "c"):
        obs_path = f'{"g2" if self.getToType(objFrom) == 4 else "talk"}/m'
        return self.downloadObjectForService(objId, path, obs_path)

    def downloadAlbumImage(self, oid: str, path: str):
        """Download album image."""
        obs_path = f"/album/a/download.nhn?ver=1.0&oid={oid}"
        r = self.server.getContent(
            self.LINE_OBS_DOMAIN + obs_path, headers=self.server.timelineHeaders
        )
        with open(path, "wb") as f:
            f.write(r.content)
        return r.content

    def downloadAlbumImageV2(
        self, _from: str, albumId: str, oid: str, savePath: str = None
    ):
        """Download album image v2."""
        ext = {"X-Line-Album": str(albumId), "X-Line-Mid": _from}
        return self.downloadObjectForService(
            oid,
            savePath,
            "album/a",
            issueToken4ChannelId="1375220249",
            additionalHeaders=ext,
        )

    def downloadProfileImage(self, mid: str, base_path: str, video: bool = False):
        url = self.LINE_OBS_DOMAIN + f"/r/talk/p/{mid}"
        savePath = f"{base_path}/{mid}.jpg"
        if video:
            url += "/vp"
            savePath = f"{base_path}/{mid}.mp4"
        r = self.server.getContent(url, headers=self.server.timelineHeaders)
        with open(savePath, "wb") as f:
            f.write(r.content)
        return savePath

    def downloadProfileCover(self, mid: str, base_path: str, video: bool = False):
        url, vc_url, objId, vc_objId = self.getProfileCoverObjIdAndUrl(mid)
        savePath = f"{base_path}/{mid}.jpg"
        r = self.server.getContent(url, headers=self.server.timelineHeaders)
        with open(savePath, "wb") as f:
            f.write(r.content)
        return savePath

    def downloadObjectMyhome(self, objId, path, objFrom="h"):
        return self.downloadObjectForService(
            objId, path, obsPathPrefix=f"myhome/{objFrom}"
        )

    def downloadImageWithURL(self, url: str, path=None):
        # TEST
        savePath = path
        if savePath is None:
            savePath = f"./.image/{time.time()}.jpg"
        r = self.server.getContant(url)
        if r.status_code != 200:
            raise Exception(f"Not a picture with URL. code: {r.status_code}")
        with open(savePath, "wb") as f:
            f.write(r.raw)
        return savePath

    def downloadObjectForService(
        self,
        objId,
        savePath,
        obsPathPrefix="myhome/h",
        size=None,
        suffix=None,
        issueToken4ChannelId: str = None,
        params: dict = None,
        additionalHeaders: dict = None,
    ):
        obs_path = f"/r/{obsPathPrefix}/{objId}"
        if params is None:
            params = {}
        if size is not None:
            # eg. size = "m800x1200" or "L800x1200"or "w800"
            # must be an existing size :p
            obs_path += f"/{size}"
        if suffix is not None:
            # eg. suffix = "mp4"
            obs_path += f"/{suffix}"
        hr = self.Hraders4Obs
        hr = self.server.additionalHeaders(
            hr,
            {
                "X-LHM": "GET",
                "X-LPV": "1",
            },
        )
        if issueToken4ChannelId is not None:
            hr = self.server.additionalHeaders(
                hr,
                {
                    "X-Line-ChannelToken": self.checkAndGetValue(
                        self.approveChannelAndIssueChannelToken(issueToken4ChannelId),
                        "channelAccessToken",
                        5,
                    ),
                },
            )
        if additionalHeaders is not None:
            hr.update(additionalHeaders)
        isDebugOnly = True
        self.log(f"Starting downloadObjectForService...", isDebugOnly)
        self.log(f"obsPath: {obs_path}", isDebugOnly)
        self.log(f"params: {params}", isDebugOnly)
        self.log(f"hr: {hr}", isDebugOnly)
        try:
            r = self.postPackDataAndGetUnpackRespData(
                f"/oa{obs_path}", b"", 0, 0, headers=hr, conn=self.obsConn
            )
        except Exception as e:
            print("downloadObjectForService: ", e)
            raise Exception(e)
        if r.status_code != 200:
            raise Exception(
                f"[downloadObjectForService] Download failed! Resp status expected `200`, got `{r.status_code}`."
            )
        if savePath is not None:
            with open(savePath, "wb") as f:
                f.write(r.content)
        self.log(f"Ended downloadObjectForService!", isDebugOnly)
        return r.content

    def uploadObjectForService(
        self,
        pathOrBytes: any,
        oType: str = "image",
        obsPath: str = "myhome/h",
        issueToken4ChannelId: str = None,
        params: dict = None,
        talkMeta: str = None,
        filename: str = None,
        additionalHeaders: dict = None,
    ):
        obs_path = f"/r/{obsPath}"
        hr = self.Hraders4Obs
        data = None
        files = None
        oType = oType.lower()
        isDebugOnly = True
        if isinstance(pathOrBytes, str):
            if pathOrBytes.startswith("http"):
                # URL
                with httpx.Client() as hc:
                    data = hc.get(pathOrBytes).content
            else:
                # FILE
                files = {"file": open(pathOrBytes, "rb")}
                filename = files["file"].name if filename is None else filename
                filename = filename.split("\\")[-1]
                with open(pathOrBytes, "rb") as f:
                    data = f.read()
        else:
            # BYTES
            data = pathOrBytes
        filename = str(uuid.uuid1()) if filename is None else filename
        base_params = {
            "type": oType,
            "ver": "2.0",
            "name": filename,
        }
        if params:
            base_params.update(params)
        params = base_params
        if files is not None:
            # FORM DATA
            data = {"params": json.dumps(params)}
            params = None
        else:
            # POST DATA
            files = None
            if len(data) == 0:
                raise ValueError("No data to send: %s" % data)
            hr = self.server.additionalHeaders(
                hr,
                {
                    "Content-Type": "image/gif",  # but...
                    "Content-Length": str(len(data)),
                    "X-Obs-Params": self.genOBSParams(params, "b64"),
                },
            )
        if issueToken4ChannelId is not None:
            hr = self.server.additionalHeaders(
                hr,
                {
                    "X-Line-ChannelToken": self.checkAndGetValue(
                        self.approveChannelAndIssueChannelToken(issueToken4ChannelId),
                        "channelAccessToken",
                        5,
                    ),
                },
            )
        if params is not None:
            hr = self.server.additionalHeaders(
                hr,
                {
                    "X-Obs-Params": self.genOBSParams(params, "b64").decode(),
                },
            )
        if talkMeta is not None:
            hr = self.server.additionalHeaders(hr, {"X-Talk-Meta": talkMeta})
        if additionalHeaders is not None:
            hr.update(additionalHeaders)
        self.log(f"Starting uploadObjectForService...", isDebugOnly)
        self.log(f"data: {str(data)[:200]}", isDebugOnly)
        self.log(f"files: {files}", isDebugOnly)
        self.log(f"obsPath: {obs_path}", isDebugOnly)
        self.log(f"params: {params}", isDebugOnly)
        self.log(f"files: {files}", isDebugOnly)
        self.log(f"hr: {hr}", isDebugOnly)
        r = self.postPackDataAndGetUnpackRespData(
            f"/oa{obs_path}",
            data,
            0,
            0,
            headers=hr,
            expectedRespCode=[200, 201],
            conn=self.obsConn,
            files=files,
        )
        # r = self.obsConn.post(self.LINE_GW_HOST_DOMAIN + f'/oa{obs_path}', data=data, headers=hr, files=files)
        # expectedRespCode:
        # - 200: image cache on obs server
        # - 201: image created success
        objId = r.headers["x-obs-oid"]
        objHash = r.headers["x-obs-hash"]  # for view on cdn
        self.log(f"Ended uploadObjectForService!", isDebugOnly)
        return objId, objHash, r.headers

    def forwardObjectForService(
        self,
        data: dict,
        contentType: str = "image",
        copyFromService: str = "talk",
        copyFromSid: str = "m",
        copyFromObjId: str = None,
        copyFrom: str = None,
        copy2Service: str = "talk",
        copy2Sid: str = "m",
        copy2ObjId: str = None,
        copy2: str = None,
        contentId: str = None,
        original: bool = False,
        duration: int = None,
        issueToken4ChannelId: str = None,
        talkMeta: str = None,
        additionalHeaders: dict = None,
    ):
        if contentType.lower() not in ["image", "video", "audio", "file"]:
            raise Exception("Type not valid.")
        if copyFrom is None:
            copyFrom = f"/{copyFromService}/{copyFromSid}"
            if copyFromObjId is not None:
                copyFrom += f"/{copyFromObjId}"
        if copy2 is None:
            copy2 = f"{copy2Service}/{copy2Sid}"
            if copy2ObjId is not None:
                copy2 += f"/{copy2ObjId}"
        base_data = {
            "copyFrom": copyFrom,
            "type": contentType,
            "ver": "1.0",
        }
        if contentId is not None:
            base_data["contentId"] = contentId
        if original:
            base_data["cat"] = "original"
        if duration is not None:
            base_data["duration"] = duration
        base_data.update(data)
        data = base_data
        hr = self.Hraders4Obs
        if issueToken4ChannelId is not None:
            hr = self.server.additionalHeaders(
                hr,
                {
                    "X-Line-ChannelToken": self.checkAndGetValue(
                        self.approveChannelAndIssueChannelToken(issueToken4ChannelId),
                        "channelAccessToken",
                        5,
                    ),
                },
            )
        if talkMeta is not None:  # Support Multiple Image
            hr = self.server.additionalHeaders(hr, {"X-Talk-Meta": talkMeta})
        if additionalHeaders is not None:
            hr.update(additionalHeaders)
        print(f"data: {data}")
        print(f"hr: {hr}")
        print(f"path: /oa/r/{copy2}/copy.nhn")
        data = list(urllib.parse.urlencode(data).encode())
        r = self.postPackDataAndGetUnpackRespData(
            f"/oa/r/{copy2}/copy.nhn", data, 0, 0, headers=hr, conn=self.obsConn
        )
        objId = r.headers["x-obs-oid"]
        objHash = r.headers["x-obs-hash"]
        return objId, objHash, r.headers

    def getObjPlayback(self, oid: int, type: str = "VIDEO"):
        url = self.LINE_HOST_DOMAIN + "/oa/talk/m/playback.obs?p=sg-1"
        data = {
            "networkType": "Cell",
            "oid": oid,
            "type": type,
            "ver": "1.0",
            "modelName": "ios",
            "lang": "zh",
        }
        hr = self.server.additionalHeaders(
            self.server.timelineHeaders,
            {
                "x-obs-channeltype": "legy",
            },
        )
        r = self.server.postContent(url, data=data, headers=hr)
        if not self.checkRespIsSuccessWithLpv(r):
            raise Exception(f"GetObjPlayback failure: {r.status_code}")
        return r.json()

    def downloadKeepContentByOidAndContentId(
        self, oid: str, contentId: str, save_path: str = None, sid: str = "p"
    ):
        return self.downloadObjectForService(
            oid,
            save_path,
            obsPathPrefix=f"keep/{sid}",
            issueToken4ChannelId="1433572998",
            params={"contentId": contentId},
        )

    def copyTalkObject4Keep(self, messageId: str, service="talk"):
        url = self.LINE_HOST_DOMAIN + f"/oa/keep/p/copy.nhn"
        hstr = "DearSakura_%s_%s" % int(time.time() * 1000), int(messageId)
        file_name = hashlib.md5(hstr.encode()).hexdigest()
        keep_oid = f"linekeep_{file_name}_tffffffff"
        data = {"oid": keep_oid, "copyFrom": f"/{service}/m/{messageId}"}
        hr = self.server.additionalHeaders(
            self.Hraders4Obs,
            {
                "x-obs-channeltype": "legy",
            },
        )
        r = self.server.postContent(url, data=data, headers=hr)
        if not self.checkRespIsSuccessWithLpv(r):
            raise Exception(f"saveTalkObject2Keep failure: {r.status_code}")
        return r.json()



# -*- coding: utf-8 -*-
"""
Automatically generated by DeachSword-DearSakura
Author: YinMo (WEDeach)
Version: 1.0.5
"""


class InterlockService(object):
    InterlockService_REQ_TYPE = 4  # BASE_VAL
    InterlockService_RES_TYPE = 4  # BASE_VAL
    InterlockService_API_PATH = None  # BASE_VAL

    def __init__(self):
        self.InterlockService_API_PATH = self.LINE_EXTERNAL_INTERLOCK_ENDPOINT

    def getPlaceSearchProviderList(self, region: str):
        """
        Get place search provider list.

        GENERATED BY YinMo0913_DeachSword-DearSakura_v1.0.5.py
        DATETIME: 02/03/2023, 23:02:07
        """
        METHOD_NAME = "getPlaceSearchProviderList"
        params = [[11, 1, region]]
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.InterlockService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.InterlockService_API_PATH, sqrd, self.InterlockService_RES_TYPE
        )

    def getPlaceSearchInfo(
        self,
        provider: int,
        keyword: str,
        clientLocale: str,
        latitude: float,
        longitude: float,
        radius: int,
    ):
        """
        Get place search info.

        GENERATED BY YinMo0913_DeachSword-DearSakura_v1.0.5.py
        DATETIME: 02/03/2023, 23:02:07
        """
        METHOD_NAME = "getPlaceSearchInfo"
        params = [
            [8, 2, provider],  # 0
            [11, 3, keyword],  # 台北101
            [11, 4, clientLocale],
            [4, 5, latitude],  # 25.03303
            [4, 6, longitude],  # 121.562759
            [8, 7, radius],
        ]
        sqrd = self.generateDummyProtocol(
            METHOD_NAME, params, self.InterlockService_REQ_TYPE
        )
        return self.postPackDataAndGetUnpackRespData(
            self.InterlockService_API_PATH, sqrd, self.InterlockService_RES_TYPE
        )
